#!/usr/bin/perl -w
#
# shinGETsu - P2P anonymous BBS
#
# (c) Fuktommy <fuktommy@users.sourceforge.net>
# Released under the GNU General Public License
# $Id: gateway.cgi,v 1.63 2004/03/01 14:43:35 fuktommy Exp $
#

use lib qw(..);
use Shingetsu::Config;
use Shingetsu::NodeList;
use Shingetsu::Cache;
use Shingetsu::CacheStat;
use Shingetsu::Signature;
use Shingetsu::Util;

umask $Shingetsu::Config::umask;

$datadir	= $Shingetsu::Config::datadir;
$menufile	= $Shingetsu::Config::menufile;
$logo		= "\xE6\x96\xb0\xE6\x9c\x88";
$anonymous	= "\xE5\x90\x8D\xE7\x84\xA1\xe3\x81\x97\xE3\x81\x95\xE3\x82\x93";

#
# sub rutines
#
sub xlocaltime(;$) { Shingetsu::Util::localtime(@_) }
sub xdie(;@) {
	print "Content-Type: text/plain;\r\n",
	      "X-Shingetsu: shinGETsu/$Shingetsu::Config::version\r\n",
	      "\r\n",
	      @_, "\n";
	exit;
}

sub str_encode($) {		# from YukiWiki
	my ($s) = @_;
	my $encoded = '';
	foreach my $ch (split(//, $s)) {
		if ($ch =~ /[A-Za-z0-9_.\/#]/) {
			$encoded .= $ch;
		} else {
			$encoded .= '%' . sprintf("%02X", ord($ch));
		}
	}
	return $encoded;
}
sub str_decode($) {		# from YukiWiki
	my ($s) = @_;
	$s =~ tr/+/ /;
	$s =~ s/%([A-Fa-f0-9][A-Fa-f0-9])/pack("C", hex($1))/eg;
	return $s;
}
sub file_encode($$) {		# from YukiWiki
	my ($type, $key) = @_;
	my $enkey = '';
	foreach my $ch (split(//, $key)) {
		$enkey .= sprintf("%02X", ord($ch));
	}
	return "${type}_${enkey}";
}
sub file_decode($) {
	my($name) = @_;
	@_ = split /_/, $name, 2;
	$name = $_[1];
	$name =~ s/[0-9A-F][0-9A-F]/pack("C", hex($&))/eg;
	return $name;
}
sub rec($) {
	return Shingetsu::Util::rec($_[0]);
}


#
# main routine
#

if ($ENV{REMOTE_ADDR} !~ $Shingetsu::Config::friendAddr) {
	print403();
	exit;
} elsif (defined $ENV{PATH_INFO}) {
	$path = $ENV{PATH_INFO};
	$path =~ s|^/||;
} else {
	$path = "";
}

if ($ENV{REQUEST_METHOD} eq "POST") {
	read(STDIN, $input, $ENV{CONTENT_LENGTH});
} else {
	$input = $ENV{QUERY_STRING};
}

if (($path eq "")||($path eq "title")) {
	printTitle();

} elsif ($path eq "list") {
	printList($menufile);

} elsif ($path eq "motd") {
	printMotd();

} elsif ($path =~ m|^list/([0-9A-Za-z_%.]+)$|) {
	printList($1);

} elsif ($path =~ m|^thread/([0-9A-Za-z_%.]+)$|) {
	printThread($1);

} elsif ($path =~ m|^attach/([0-9A-Za-z_%.]+)/([0-9A-Za-z_]+)/([0-9]+)\.([0-9A-Za-z_]*)$|) {
	printAttach($1, $2, $3, $4);

} elsif ($path eq "check") {
	checkSign();

} elsif ($path eq "new") {
	printHeader("new");
	print "<h1>new</h1>";
	newElementForm();
	print "</body></html>\n";

} elsif ($path eq "index") {
	checkAdmin();
	printIndex();

} elsif ($path eq "changes") {
	checkAdmin();
	printChanges();

} elsif ($path eq "update") {
	checkAdmin();
	printUpdate();

} elsif ($path eq "search") {
	checkAdmin();
	if ((defined $input) && ($input ne "")) {
		%arg = args($input);
		printSearchResult($arg{query});
	} else {
		printHeader("search");
		print "<h1>Search</h1>";
		searchForm();
		print "</body></html>";
	}

} elsif ($path eq "edittrust") {
	checkAdmin();
	%arg = args($input);
	if (! defined $arg{trip}) {
		print404();
		exit;
	} elsif ($arg{button} eq "ADD") {
		Shingetsu::Signature::addlist($arg{trip});
	} elsif ($arg{button} eq "DELETE") {
		Shingetsu::Signature::dellist($arg{trip});
	}
	print302("$ENV{SCRIPT_NAME}/$arg{type}/" .
		str_encode(file_decode("$arg{type}_$arg{file}")));


} elsif ($path =~ m|^post/([0-9A-Za-z_%.]+)$|) {
	my $num = $1;
	my $id = post($num, $input);
	print302("$ENV{SCRIPT_NAME}/thread/" . str_encode(file_decode($num)) . "#r$id");

} elsif ($path eq "newelement") {
	%arg = args($input);
	xdie "Null Name" if ($arg{name} eq "");
	xdie "Bad Name" if ($arg{name} =~ m|[/#]|);
	xdie "Null Type" if ($arg{type} eq "");
	print302("$ENV{SCRIPT_NAME}/$arg{type}/" . str_encode($arg{name}));

} elsif ($path =~ m|^newelement/([0-9A-Za-z_]+)$|) {
	my $num = $1;
	newElement($num, $input);
	print302("$ENV{SCRIPT_NAME}/list/" . str_encode(file_decode($num)));

} elsif ($path =~ m|^delete/([0-9A-Za-z_]+)$|) {
	checkAdmin();
	my $file = $1;
	deleteDialog($file);

} elsif ($path =~ m|^xdelete/([0-9A-Za-z_]+)$|) {
	checkAdmin();
	my $file = $1;
	deleteFile($file);

} elsif ($path =~ m|^xdelete/([0-9A-Za-z_]+)/([0-9]+)/([0-9A-Za-z_]+)$|) {
	my($file, $stamp, $id) = ($1, $2, $3);
	checkAdmin();
	deleteRecord($file, $stamp, $id);

} else {
	print404();

}

#------------------------------------------------------------------#

#
# check admin address
#
sub checkAdmin {
	if ($ENV{REMOTE_ADDR} !~ $Shingetsu::Config::adminAddr) {
		print403();
		exit;
	}
}

#
# print CGI&HTML header
#
sub printHeader {
	my($title) = @_;
	my $ver = $Shingetsu::Config::version;
	print "Content-Type: text/html; charset=UTF-8\r\n",
	      "Content-Language: ja\r\n",
	      "X-Shingetsu: shinGETsu/$ver\r\n",
	      "\r\n";
	print '<?xml version="1.0" encoding="UTF-8"?>',
	      '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"',
	      ' "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">',
	      '<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="ja">',
	      '<head>',
	      '<meta http-equiv="Content-Language" content="ja" />',
	      '<meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />',
	      "<meta http-equiv='Content-Style-Type' content='text/css' />",
	      "<title>$title</title>",
	      "<link rev='made' href='mailto:fuktommy\@users.sourceforge.net' />",
	      "<link rel='index' href='$ENV{SCRIPT_NAME}/list' />",
	      "<style type='text/css'>",
	      " ul { list-style: none; }",
	      " h1 { color: red; }",
	      " em { font-size: x-large }",
	      " .name { color: green; font-weight: bold; }",
	      "</style></head>";
	print "<body><p>",
	      "<a href='$ENV{SCRIPT_NAME}/'>TOP</a> | ",
	      "<a href='$ENV{SCRIPT_NAME}/list'>MENU</a> | ",
	      "<a href='$ENV{SCRIPT_NAME}/new'>NEW</a> | ",
	      "<a href='$ENV{SCRIPT_NAME}/index'>INDEX</a> | ",
	      "<a href='$ENV{SCRIPT_NAME}/changes'>CHANGES</a> | ",
	      "<a href='$ENV{SCRIPT_NAME}/update'>UPDATE</a> | ",
	      "<a href='$ENV{SCRIPT_NAME}/search'>SEARCH</a>",
	      "</p>";
}

#
# print CGI header (403 forbidden)
#
sub print403 {
	my ($file) = @_;
	printHeader("403 Forbidden");
	print "<h1>You are not the administrator.</h1>",
	      "</body></html>\n";
}

#
# print CGI header (404 not found)
#
sub print404 {
	my ($file) = @_;
	printHeader("404 Not Found");
	print "<h1>Bad arguments or No data.</h1><p>Try later.</p>";
	if (defined $file) {
	      print "<form method='get' action='$ENV{SCRIPT_NAME}/delete/$file'>",
		    "<p><input type='submit' value='DELETE FILE' tabindex='1' accesskey='d' /></p></form>";
	}
	print "</body></html>\n";
}

#
# print CGI header (302 moved temporarily)
#
sub print302 {
	my($next) = @_;
	my $ver = $Shingetsu::Config::version;
	my($title) = @_;
	print "Content-Type: text/html; charset=UTF-8\r\n",
	      "Content-Language: ja\r\n",
	      "X-Shingetsu: shinGETsu/$ver\r\n",
	      "\r\n";
	print '<?xml version="1.0" encoding="UTF-8"?>',
	      '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"',
	      ' "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">',
	      '<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="ja">',
	      '<head>',
	      '<meta http-equiv="Content-Language" content="ja" />',
	      '<meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />',
	      '<meta http-equiv="Content-Script-Type" content="text/javascript" />',
	      "<title>moved</title>",
	      "<link rev='made' href='mailto:fuktommy\@users.sourceforge.net' />",
	      "<link rel='index' href='$ENV{SCRIPT_NAME}/list' />",
	      "</head>";
	print "<body><p>Click and jump to <a href='$next'>$next</a></p>\n",
	      "<script type='text/javascript'>window.location.href = '$next';</script>",
	      "</body></html>\n";
}

#
# print title page
#
sub printTitle {
	printHeader($logo);
	print "<h1>$logo - P2P anonymous BBS</h1>",
	      "<p>shinGETsu is released under the GNU General Public License</p>\n",
	      "<ul><li><em><a href='$ENV{SCRIPT_NAME}/list'>MENU</a></em></li>\n";
	print "<li><a href='/tradgw.cgi'>",
	      "Traditional Style Gateway</a></li>\n" if (-x "tradgw.cgi");
	print "<li><a href='http://shingetsu.sourceforge.net/'>",
	      "Official Site</a></li>\n",
	      "<li><a href='$ENV{SCRIPT_NAME}/motd'>Agreement</a></li></ul>",
	      "</body></html>\n";
}

#
# print MOTD file
#
sub printMotd {
	print "Content-type: text/plain; charset=euc-jp\r\n",
	      "X-Shingetsu: shinGETsu/$Shingetsu::Config::version\r\n\r\n";
	open IN, $Shingetsu::Config::motd;
	while (<IN>) {
		print;
	}
	close IN;
}

#
# print list
#
sub printList {
	my($list) = @_;
	local($_, @_);
	$list = str_decode($list);
	my $file_list = file_encode("list", $list);
	touch($file_list);
	open IN, "$datadir/$file_list.dat" or die "$datadir/$file_list.dat: failed to open. $!";
	printHeader($list);
	print "<h1>$list</h1>",
	      "<form method='get' action='$ENV{SCRIPT_NAME}/delete/$file_list'><ul>\n";
	my @buf = ();
	my %stamp = ();
	my %stat = Shingetsu::CacheStat::list();
	my $count = 0;
	while (<IN>) {
		$count++;
		my %rec = rec($_);
		next if (keys %rec <= 2);
		$rec{body} = "" unless (defined $rec{body});
		my $stamp = $rec{stamp};
		my $buf;
		if ($rec{type}) {
			my $file_child = file_encode($rec{type}, $rec{link});
			my $file = "$datadir/$file_child.dat";
			my $records = 0;
			if (-s $file) {
				$stamp   = $stat{$file_child}{stamp};
				$records = $stat{$file_child}{records};
			}
			$records = "?" if ($records <= 0);
			my $str = xlocaltime($stamp);
			my $child = str_encode($rec{link});
			$buf = "<li><input type='radio' name='record' value='$rec{stamp}/$rec{id}' " .
			       "tabindex='1' accesskey='s' />" .
			       "$str: <a href='$ENV{SCRIPT_NAME}/$rec{type}/$child'>$rec{link}</a>" . 
			       "($rec{type}: $records)$rec{body}</li>\n";
		} elsif ($rec{remove_stamp}) {
			$stamp = $rec{remove_stamp};
			my $str = xlocaltime($stamp);
			my $check = "";
			$rec{name} = "" unless (defined $rec{name});
			if ($rec{pubkey}) {
				$check = "<a href='$ENV{SCRIPT_NAME}/check?" .
					"file=" . file_encode("list", $list) . "&amp;" .
					"stamp=$rec{stamp}&amp;id=$rec{id}'>" .
					Shingetsu::Signature::pubkey2trip($rec{pubkey}) . "</a>";
			}
			$buf = "<li><input type='radio' name='record' value='$rec{stamp}/$rec{id}' " .
			       "tabindex='1' accesskey='s' />" .
			       "$str: $rec{name} $check remove:$rec{remove} $rec{body}</li>\n";
		}
		push @buf, $buf;
		$stamp{$buf} = $stamp;
	}
	close IN;
	@buf = sort {$stamp{$b} <=> $stamp{$a}} @buf;
	print @buf;
	print "</ul><p><input type='submit' value='DELETE RECORD' tabindex='2' accesskey='d' />",
	      "<input type='hidden' name='mode' value='list' /></p></form>";
	my $filesize = int((-s "$datadir/$file_list.dat")/1024/1024*10)/10;
	my $checked = ($count)? "checked='checked'": "";
	if ($filesize <= $Shingetsu::Config::filelimit) {
		newElementForm(file=>$file_list, checked=>$checked, dopost=>1);
	}
	print "<form action='$ENV{SCRIPT_NAME}/delete/$file_list'>",
	      "<p><input type='submit' value='DELETE FILE' tabindex='6' accesskey='d' />",
	      " ${filesize}MB</p></form></body></html>\n";
}

#
# print thread
#
sub printThread {
	my($thread) = @_;
	local $_;
	$thread = str_decode($thread);
	my $file_thread = file_encode("thread", $thread);
	touch($file_thread);
	open IN, "$datadir/$file_thread.dat" or die "$datadir/$file_thread.dat: failed to open. $!";
	printHeader($thread);
	print "<h1>$thread</h1>",
	      "<form method='get' action='$ENV{SCRIPT_NAME}/delete/$file_thread'><dl>\n";
	my $count = 0;
	while (<IN>) {
		$count++;
		%rec = rec($_);
		next if (keys %rec <= 2);
		my $xstamp = xlocaltime $rec{stamp};
		$rec{name} = $anonymous if ((! defined $rec{name}) || ($rec{name} eq ""));
		$rec{mail} = "" if (! defined $rec{mail});
		$rec{body} = "" if (! defined $rec{body});
		$rec{body} =~ s|(&gt;&gt;)([0-9A-Za-z]+)|<a href='#r$2'>$&</a>|g;
		$rec{body} =~ s|https?://[\041-\073\075-\132\136-\177]{2,}|<a href='$&'>$&</a>|g;
		$rec{body} =~ s|\[\[<a.*?>(.*?)\]\]</a>|[[$1]]|g;
		$rec{body} =~ s|\[\[([^<>]+?)\]\]|sprintf("<a href='%s'>[[%s]]</a>",str_encode($1),$1)|eg;
		if ($Shingetsu::Config::redirect) {
			$rec{body} =~ s|<a href='http://shingetsu1\.p2p/gateway\.cgi|<a href='/tradgw.cgi|g if (-x "tradgw.cgi");
			$rec{body} =~ s|<a href='http://shingetsu2\.p2p/|<a href='/|g;
		}
		$rec{body} =~ s|<BR>|<br />|ig;
		my $attach = "";
		my $check  = "";
		my $remove = "";
		if ($rec{attach}) {
			my $size = int(length($rec{attach}) / 4 * 3 / 1024);
			$attach = " <a href='$ENV{SCRIPT_NAME}/attach/" .
				  str_encode($thread) . "/$rec{id}/$rec{stamp}.$rec{suffix}'>" .
				  "$rec{stamp}.$rec{suffix}</a> (${size}KB)";
		}
		if ($rec{pubkey}) {
			$check = "<a href='$ENV{SCRIPT_NAME}/check?" .
				 "file=" . file_encode("thread", $thread) . "&amp;" .
				 "stamp=$rec{stamp}&amp;id=$rec{id}'>" .
				 Shingetsu::Signature::pubkey2trip($rec{pubkey}) . "</a>";
		}
		if (($rec{remove_id}) && ($rec{remove_stamp})) {
			my $xid = substr($rec{remove_id},0,8);
			$remove = "[[remove: <a href='#r$xid'>$xid</a>]]";
		}
		my $xid = substr $rec{id}, 0, 8;
		print "<dt id='r$xid'><input type='radio' name='record' value='$rec{stamp}/$rec{id}' ",
		      "tabindex='1' accesskey='s' />",
		      "$xid :<span class='name'>$rec{name}</span>",
		      " [$rec{mail}] $check $xstamp$attach</dt><dd>$rec{body}<br />$remove<br /></dd>\n";
	}
	close IN;
	print "</dl><p><input type='submit' value='DELETE RECORD' tabindex='2' accesskey='d' />",
      	      "<input type='hidden' name='mode' value='thread' /></p></form>";
	my $filesize = int((-s "$datadir/$file_thread.dat")/1024/1024*10)/10;
	my $checked = ($count)? "checked='checked'": "";
	my $readonly = ($ENV{REMOTE_ADDR} =~ $Shingetsu::Config::adminAddr)? "": "readonly='readonly'";
	if ($filesize <= $Shingetsu::Config::filelimit) {
		print "<form method='post' action='$ENV{SCRIPT_NAME}/post/$file_thread' enctype='multipart/form-data'><p>",
		"<input type='submit' value='POST' name='submit' tabindex='1' accesskey='w' />",
		" <input type='checkbox' value='dopost' name='dopost'",
		" $checked tabindex='2' accesskey='s' />",
		"Send to other nodes",
		" <input type='checkbox' value='error' name='error'",
		" checked='checked' tabindex='2' accesskey='e' />",
		"Error in timestamp<br />",
		" Name:<input name='name' size='15' value='' tabindex='3' accesskey='n' />",
		" E-mail:<input name='mail' size='15' value='' tabindex='4' accesskey='m' />",
		" Trip:<input type='password' name='passwd' size='15' $readonly value='' tabindex='5' accesskey='p' /><br />",
		" Attach:<input type='file' name='attach' size='19' value='' tabindex='6' accesskey='a' />",
		" Type:<select name='suffix' size='1' tabindex='7'>",
		"<option>AUTO</option>";
		foreach (sort (keys %Shingetsu::Config::mimeType)) {
			print "<option>$_</option>";
		}
		print "</select> (limit: ${Shingetsu::Config::filelimit}MB)<br />",
		"<textarea rows='5' cols='70' name='message' tabindex='8' accesskey='c'></textarea>",
		"</p></form>";
	}
	print "<form method='get' action='$ENV{SCRIPT_NAME}/delete/$file_thread'>",
	      "<p><input type='submit' value='DELETE FILE' tabindex='9' accesskey='d' />",
	      " ${filesize}MB</p></form></body></html>\n";
}

#
# print attachment
#
sub printAttach {
	my($thread, $id, $stamp, $suffix) = @_;
	$thread = file_encode("thread", str_decode($thread));
	my $ver = $Shingetsu::Config::version;
	$suffix = lc $suffix;
	my $type = $Shingetsu::Config::mimeType{$suffix};
	$type = "text/plain" unless (defined $type);
	touch($thread);
	open IN, "$datadir/$thread.dat" or die "$datadir/$thread.dat: failed to open. $!";
	my $value;
	while (<IN>) {
		my %rec = rec($_);
		if ($rec{attach} && ($rec{stamp} eq $stamp) && ($rec{id} eq $id)) {
			$value = $rec{attach};
			last;
		}
	}
	close IN;
	if (defined $value) {
		print "Content-Type: $type\r\n",
		      "X-Shingetsu: shinGETsu/$ver\r\n",
		      "\r\n",
		       Shingetsu::Extern::base64decode($value);
	} else {
		print404($thread);
	}
}

#
# check signature
#
sub checkSign {
	my %arg = args($input);
	unless ($arg{file} && $arg{stamp} && $arg{id}) {
		print404();
		exit;
	}
	my $file = $arg{file};
	touch($file);
	open IN, "$datadir/$file.dat" or die "$datadir/$file.dat: failed to open. $!";
	my $value;
	while (<IN>) {
		my %rec = rec($_);
		next unless (($rec{stamp} eq $arg{stamp}) && ($rec{id} eq $arg{id}));
		if (! ((defined $rec{pubkey}) && ($rec{pubkey} ne "")
			&& (defined $rec{sign}) && ($rec{sign} ne "")
			&& (defined $rec{target}) && ($rec{target} ne ""))) {
			print "Content-Type: text/plain; charset=UTF-8\r\n",
			      "X-Shingetsu: shinGETsu/$Shingetsu::Config::version\r\n",
			      "\r\n",
			      "This record is not signed.\n$_\n";
		} elsif (Shingetsu::Signature::check(%rec)) {
			s/</&lt;/g; s/>/&gt;/g;
			my $trip = Shingetsu::Signature::pubkey2trip($rec{pubkey});
			my($type, $filename) = split /_/, $file, 2;
			printHeader("check sign");
			print "<p>This record is protected($rec{target}).</p>",
			"<form method='get' action='$ENV{SCRIPT_NAME}/edittrust'><p>",
			"<input type='submit' name='button' value='ADD' tabindex='1' accesskey='a' />",
			"/<input type='submit' name ='button' value='DELETE' tabindex='2' accesskey='d' />",
			"<input type='hidden' name='trip' value='$trip' />",
			"<input type='hidden' name='type' value='$type' />",
			"<input type='hidden' name='file' value='$filename' />",
			"this trip to/from trust list: $trip</p></form>",
			"<p>$_</p>",
			"</body></html>";
			close IN;
			exit;
		} else {
			print "Content-Type: text/plain; charset=UTF-8\r\n",
			      "X-Shingetsu: shinGETsu/$Shingetsu::Config::version\r\n",
			      "\r\n",
			      "This record has WRONG sign.\n$_\n";
			close IN;
			exit;
		}
	}
	close IN;
	print404($file);
}


#
# make file
#
sub touch {
	my($file) = @_;
	my %stat = Shingetsu::CacheStat::list();
	unless (-s "$datadir/$file.dat") {
		Shingetsu::Util::lock();
		open OUT, ">> $datadir/$file.dat";
		close OUT;
		Shingetsu::Util::unlock();
		my %stat = Shingetsu::CacheStat::list();
		$stat{$file}{records} = 0;
		$stat{$file}{stamp}   = 0;
		Shingetsu::CacheStat::sync(%stat);
	}
	if ((-s "$datadir/$file.dat") == 0) {
		my $node = Shingetsu::Cache::search($file);
		if (defined $node) {
			Shingetsu::Cache::getRegion($file, $node);
		}
	}
}

#
# print index page
#
sub printIndex {
	printHeader("index");
	print "<h1>Index</h1>\n";
	local $_;
	my $type;
	my %stat = Shingetsu::CacheStat::list();
	foreach $type ("list", "thread") {
		print "<h2>$type</h2><ul>\n";
		foreach (glob("$datadir/${type}_*.dat")) {
			$_ =~ s|^$datadir/||;	$_ =~ s|.dat$||;
			my $x = file_decode($_);
			my $y = str_encode($x);
			my $size = $stat{$_}{records};
			print "<li><a href='$ENV{SCRIPT_NAME}/$type/$y'>$x</a>($size)</li>\n";
		}
		print "</ul>"
	}
	newElementForm();
	print "</body></html>\n";
}

#
# print recent changes
#
sub printChanges {
	printHeader("changes");
	print "<h1>Changes</h1>\n";
	local $_;
	my $type;
	my %stat = Shingetsu::CacheStat::list();
	foreach $type ("list", "thread") {
		print "<h2>$type</h2><ul>\n";
		my @buf = ();
		my %stamp = ();
		foreach (glob("$datadir/${type}_*.dat")) {
			$_ =~ s|^$datadir/||;	$_ =~ s|.dat$||;
			my $x = file_decode($_);
			my $y = str_encode($x);
			my $date = xlocaltime($stat{$_}{stamp});
			my $size = $stat{$_}{records};
			my $buf = "<li>$date <a href='$ENV{SCRIPT_NAME}/$type/$y'>$x</a>($size)</li>\n";
			$stamp{$buf} = $stat{$_}{stamp};
			push @buf, "$buf";
		}
		@buf = sort {$stamp{$b} <=> $stamp{$a}} @buf;
		print @buf;
		print "</ul>"
	}
	newElementForm();
	print "</body></html>\n";
}

#
# print recent update
#
sub printUpdate {
	local $_;
	my %stamp = ();
	if (open IN, $Shingetsu::Config::updatelist) {
		while (<IN>) {
			chomp;
			@_ = split /<>/;
			if ($_[2] =~ /^(list|thread)_/) {
				$stamp{$_[2]} = $_[0];
			}
		}
		close IN;
	} else {
		print404("update");
		exit;
	}
	printHeader("update");
	print "<h1>Update</h1>\n";
	my $type;
	my %stat = Shingetsu::CacheStat::list();
	foreach $type ("list", "thread") {
		print "<h2>$type</h2><ul>\n";
		my @buf = ();
		foreach (sort {$stamp{$b}<=>$stamp{$a}} grep(/^${type}_/, keys(%stamp))) {
			my $x = file_decode($_);
			my $y = str_encode($x);
			my $date = xlocaltime $stamp{$_};
			my $size = (defined $stat{$_})? $stat{$_}{records}: "?";
			print "<li>$date <a href='$ENV{SCRIPT_NAME}/$type/$y'>$x</a>($size)</li>\n";
		}
		print "</ul>"
	}
	newElementForm();
	print "</body></html>\n";
}

#
# escape
#
sub escape {
	local($_) = @_;
	s/&/&amp;/g;
	s/&amp;(#\d+|#[Xx][0-9A-Fa-f]+|[A-Za-z0-9]+);/&$1;/g;
	s/</&lt;/g;
	s/>/&gt;/g;
	s/\r//g;
	s/\n/<br>/g;
	return $_;
}

#
# unescape arguments
#
sub args {
	local $_ = $_[0];
	local @_ = split /&/;
	my %arg = ();
	foreach (@_) {
		local @_ = split /=/;
		if (defined $_[1]) {
			$_[1] =~ s/\+/ /g;
			$_[1] =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
			$arg{$_[0]} = escape($_[1]);
		} else {
			$arg{$_[0]} = "";
		}
	}
	return %arg;
}

#
# unescape arguments(multipart/form-data)
#
sub argsFromMulti {
	my($input) = @_;
	local $_;
	my %arg = ();
	$input =~ /\s/;
	my $boundary = $`;
	my @input = split $boundary, $input;
	shift @input;
	foreach (@input) {
		my @buf = split /\r\n/, $_;
		next if (@buf == 1);
		shift @buf;
		$buf[0] =~ /Content-Disposition: form-data; name="([^"]+)"/i or next;
		my $key = $1;

		if ($buf[0] =~ /filename="([^"]+)"/i) {
			$arg{auto_suffix} = $1;
			$arg{auto_suffix} =~ s/.*[\/\\]//;
			if ($arg{auto_suffix} =~ /\.([^.]*)$/) {
				$arg{auto_suffix} = $1;
				$arg{auto_suffix} = lc $arg{auto_suffix};
			} else {
				$arg{auto_suffix} = "";
			}
		}

		$_ = shift @buf until ((! defined $_) || ($_ eq ""));
		$arg{$key} = join "\r\n", @buf;
		if ($key ne "attach") {
			$arg{$key} = escape($arg{$key});
		}
	}
	return %arg;
}

#
# post article
#
sub post {
	my($thread, $input) = @_;
	local $_;
	my %arg = argsFromMulti($input);
	my ($suffix, $attach) = ("", "");
	$arg{name} = "" unless (defined $arg{name});
	$arg{mail} = "" unless (defined $arg{mail});
	$arg{passwd} = "" unless (defined $arg{passwd});
	if ((defined $arg{attach}) && ($arg{attach} ne "")) {
		if (length $arg{attach} > $Shingetsu::Config::filelimit*1024*1024) {
			xdie "Too big file";
		}
		$attach = Shingetsu::Extern::base64encode($arg{attach});
		if ($arg{suffix} ne "AUTO") {
			$suffix	= $arg{suffix};
		} elsif ((defined $arg{auto_suffix}) && ($arg{auto_suffix} ne "")) {
			$suffix	= $arg{auto_suffix};
		} else {
			$suffix	= "txt";
		}
	}
	my $stamp = (defined $arg{error})? Shingetsu::Util::time(): time;
	my $body  = "";
	my $target = "";
	if ((defined $arg{name})&&($arg{name} ne "")) {
		$body .= "<>name:$arg{name}";
		$target .= ",name";
	}
	if ((defined $arg{mail})&&($arg{mail} ne "")) {
		$body .= "<>mail:$arg{mail}";
		$target .= ",mail";
	}
	if ((defined $arg{message})&&($arg{message} ne "")) {
		$body .= "<>body:$arg{message}";
		$target .= ",body";
	}
	if ((defined $suffix)&&($suffix ne "")) {
		$body .= "<>suffix:$suffix";
		$target .= ",suffix";
	}
	if ((defined $attach)&&($attach ne "")) {
		$body .= "<>attach:$attach";
		$target .= ",attach";
	}
	$body =~ s/^<>//;
	$target =~ s/^,//;
	xdie "Null Article" if ($body eq "");
	if ($arg{passwd} ne "") {
		my($pubkey, $sign) = Shingetsu::Signature::generate($arg{passwd}, $body);
		$body = "pubkey:$pubkey<>sign:$sign<>target:$target<>" . $body;
	}
	my $id = addRecord($thread, $stamp, $body);
	if (defined $arg{dopost}) {
		Shingetsu::Util::addUpdate($stamp, $id, $thread);
		Shingetsu::Cache::tellupdate($thread, $stamp, $id, undef);
	}
	return substr $id, 0, 8;
}

#
# make new list element
#
sub newElement {
	my($file, $input) = @_;
	local $_;
	my %arg = args($input);
	my $stamp = Shingetsu::Util::time();
	xdie "Null Name" if ($arg{name} eq "");
	xdie "Bad Name" if ($arg{name} =~ m|[/#]|);
	xdie "Null Type" if ($arg{type} eq "");
	my $body = "type:$arg{type}<>link:$arg{name}";
	my $id = addRecord($file, $stamp, $body);
	if (defined $arg{dopost}) {
		Shingetsu::Util::addUpdate($stamp, $id, $file);
		Shingetsu::Cache::tellupdate($file, $stamp, $id, undef);
	}
}

#
# delete dialog
#
sub deleteDialog {
	my($file) = @_;
	my %arg = args($input);
	local $_;
	if (defined $arg{record}) {
		printHeader("delete record");
		print "<h1>Delete Record</h1>\n";
		my($stamp, $id) = split /\//, $arg{record};
		my $buf;
		if (open IN, "$datadir/$file.dat") {
			while (<IN>) {
				if (/$stamp<>$id<>/) {
					$buf = $_;
					last;
				}
			}
			close IN;
			if (! defined $buf) {
				print "<p>no such record</p></body></html>\n";
			} else {
				$buf =~ s/</&lt;/g;
				$buf =~ s/>/&gt;/g;
				print "<p>Do you delete following record?</p>",
				"<form method='post' action='$ENV{SCRIPT_NAME}/xdelete/$file/$stamp/$id'><p>",
				"<input type='submit' value='YES' name='submit' tabindex='1' accesskey='w' />",
				" <input type='checkbox' value='dopost' name='dopost' tabindex='2' accesskey='s' />",
				"Send to other nodes",
				" <input type='checkbox' value='error' name='error'",
				" checked='checked' tabindex='2' accesskey='e' />",
				"Error in timestamp<br />",
				" Name:<input name='name' size='15' value='' tabindex='3' accesskey='n' />",
				" Trip:<input type='password' name='passwd' size='15' value='' tabindex='4' accesskey='p' />",
				" Comment:<input name='message' size='15' value='' tabindex='5' accesskey='m' />",
				"<input type='hidden' name='mode' value='$arg{mode}' />",
				"</p></form>\n",
				"<p>$buf</p></body></html>\n";
			}
		} else {
			print "<p>$file: no such file</p></body></html>\n";
		}

	} else {
		printHeader("delete file");
		print "<h1>Delete File</h1>\n";
		my @buf = ();
		my $count = 0;
		if (open IN, "$datadir/$file.dat") {
			while (<IN>) {
				push @buf, $_;
				last if ($count++ > 1);
			}
			close IN;
			print "<p>Do you delete following file? ",
			      "<a href='$ENV{SCRIPT_NAME}/xdelete/$file'>",
			      "YES</a></p>\n";
			foreach (@buf) {
				chomp;
				s/</&lt;/g;
				s/>/&gt;/g;
				print "<p>$_</p>\n";
			}
			print "</body></html>\n";
		} else {
			print "<p>$file: no such file</p></body></html>\n";
		}

	}

}

#
# delete file
#
sub deleteFile {
	my($file) = @_;
	if (unlink "$datadir/$file.dat") {
		my %stat = Shingetsu::CacheStat::list();
		delete $stat{$file};
		Shingetsu::CacheStat::sync(%stat);
		print302("$ENV{SCRIPT_NAME}/list");
	} else {
		print "Content-Type: text/plain\r\n\r\n",
		      "failed. $!\n";
	}
}

#
# delete record
#
sub deleteRecord {
	my($file, $stamp, $id) = @_;
	my %arg = args($input);
	local $_;
	if (! Shingetsu::Cache::removeRecord($file, $stamp, $id)) {
		print "Content-Type: text/plain\r\n\r\n",
		      "failed. $!\n";
	} elsif ($arg{dopost}) {
		$arg{name} = "" unless (defined $arg{name});
		$arg{mail} = "" unless (defined $arg{mail});
		$arg{passwd} = "" unless (defined $arg{passwd});
		my $newstamp = (defined $arg{error})? Shingetsu::Util::time(): time;
		my $body  = "";
		my $target = "";
		if ((defined $arg{name})&&($arg{name} ne "")) {
			$body .= "<>name:$arg{name}";
			$target .= ",name";
		}
#		if ((defined $arg{mail})&&($arg{mail} ne "")) {
#			$body .= "<>mail:$arg{mail}";
#			$target .= ",mail";
#		}
		if ((defined $arg{message})&&($arg{message} ne "")) {
			$body .= "<>body:$arg{message}";
			$target .= ",body";
		}
		$body .= "<>remove_stamp:$stamp";	$target .= ",remove_stamp";
		$body .= "<>remove_id:$id";		$target .= ",remove_id";
		$body =~ s/^<>//;
		$target =~ s/^,//;
		xdie "Null Article" if ($body eq "");
		if ($arg{passwd} ne "") {
			my($pubkey, $sign) = Shingetsu::Signature::generate($arg{passwd}, $body);
			$body = "pubkey:$pubkey<>sign:$sign<>target:$target<>" . $body ;
		}
		my $newid = addRecord($file, $newstamp, $body);
		Shingetsu::Util::addUpdate($newstamp, $newid, $file);
		Shingetsu::Cache::tellupdate($file, $newstamp, $newid, undef);
	}
	if ((defined $arg{mode}) && ($arg{mode} ne "")) {
			my $xfile = str_encode(file_decode($file));
			print302("$ENV{SCRIPT_NAME}/$arg{mode}/$xfile");
	} else {
		print "Content-Type: text/plain\r\n\r\n",
		      "succeed. $!\n";
	}
}

#
# add record
#
sub addRecord {
	my($file, $stamp, $body) = @_;
	my $id = Shingetsu::Extern::md5digest($body);
	Shingetsu::Util::lock();
	open OUT, ">> $datadir/$file.dat" or die "$datadir/$file.dat: failed to open. $!";
	print OUT "$stamp<>$id<>$body\n";
	close OUT;
	Shingetsu::Util::unlock();
	my %stat = Shingetsu::CacheStat::list();
	$stat{$file}{records}++;
	$stat{$file}{stamp} = $stamp;
	Shingetsu::CacheStat::sync(%stat);
	return $id;
}

#
# new element form
# $opt{file}, $opt{checked}, $opt{dopost}
#
sub newElementForm {
	my %opt = @_;
	$opt{checked} = "" unless ($opt{checked});
	my $file = (defined $opt{file})? "/$opt{file}": "";

	print "<form method='get' action='$ENV{SCRIPT_NAME}/newelement$file'><p>",
	      "<input type='submit' value='NEW' name='submit' tabindex='1' accesskey='w' />",
	      "<input type='radio' value='list' name='type' tabindex='2' accesskey='t' />",
	      "list",
	      " <input type='radio' value='thread' name='type' tabindex='2' accesskey='t' />",
	      "thread";
	print " <input type='checkbox' value='dopost' name='dopost'",
	      " $opt{checked} tabindex='2' accesskey='s' />",
	      "Send to other nodes" if ($opt{dopost});
	print "<br />Name: <input name='name' size='19' value='' tabindex='3' accesskey='t' />",
	      "</p></form>";
}

#
# search form
#
sub searchForm {
	my $query = (@_)? $_[0]: "";
	print "<form method='get' action='$ENV{SCRIPT_NAME}/search'><p>",
	      "<input name='query' size='19' value='$query' tabindex='1' accesskey='q' />",
	      "<input type='submit' value='SEARCH' name='submit' tabindex='1' accesskey='s' />",
	      "</p></form>";
}

#
# search result
#
sub printSearchResult {
	my($query) = @_;
	printHeader("search");
	print "<h1>Search</h1>\n";
	local $_;
	my($type, $file);
	my %stat = Shingetsu::CacheStat::list();
	foreach $type ("list", "thread") {
		print "<h2>$type</h2><ul>\n";
		foreach $file (glob("$datadir/${type}_*.dat")) {
			my $file2 = $file;
			$file2 =~ s|^$datadir/||; $file2 =~ s|.dat$||;
			my $x = file_decode($file2);
			my $y = str_encode($x);
			my $found;
			if ($x =~ /$query/) {
				$found = 1;
			} else {
				open IN, $file or die "$file: open failed. $!";
				while (<IN>) {
					if (/$query/) {
						$found = 1;
						last;
					}
				}
				close IN;
			}
			next unless ($found);
			my $size = $stat{$file2}{records};
			print "<li><a href='$ENV{SCRIPT_NAME}/$type/$y'>$x</a>($size)</li>\n";
		}
		print "</ul>"
	}
	searchForm($query);
	print "</body></html>";
}
